==> databricks-connect-14.3.9-sources.jar: org/apache/spark/sql/streaming/StreamingQueryListener.scala <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql.streaming

import java.lang.{Long => JLong}
import java.util.UUID

import scala.collection.JavaConverters._

import org.json4s._
import org.json4s.JsonAST.JValue
import org.json4s.JsonDSL._
import org.json4s.jackson.JsonMethods._

import org.apache.spark.annotation.Evolving

/**
 * Interface for listening to events related to [[StreamingQuery StreamingQueries]].
 * @note
 *   The methods are not thread-safe as they may be called from different threads.
 *
 * @since 3.5.0
 */
@Evolving
abstract class StreamingQueryListener extends Serializable {

  import StreamingQueryListener._

  /**
   * Called when a query is started.
   * @note
   *   This is called synchronously with
   *   [[org.apache.spark.sql.streaming.DataStreamWriter `DataStreamWriter.start()`]], that is,
   *   `onQueryStart` will be called on all listeners before `DataStreamWriter.start()` returns
   *   the corresponding [[StreamingQuery]]. Please don't block this method as it will block your
   *   query.
   * @since 3.5.0
   */
  def onQueryStarted(event: QueryStartedEvent): Unit

  /**
   * Called when there is some status update (ingestion rate updated, etc.)
   *
   * @note
   *   This method is asynchronous. The status in [[StreamingQuery]] will always be latest no
   *   matter when this method is called. Therefore, the status of [[StreamingQuery]] may be
   *   changed before/when you process the event. E.g., you may find [[StreamingQuery]] is
   *   terminated when you are processing `QueryProgressEvent`.
   * @since 3.5.0
   */
  def onQueryProgress(event: QueryProgressEvent): Unit

  /**
   * Called when the query is idle and waiting for new data to process.
   * @since 3.5.0
   */
  def onQueryIdle(event: QueryIdleEvent): Unit = {}

  /**
   * Called when there is an asynchronous state commit complete for a microbatch in the query. This
   * event is generated only when asynchronous state checkpointing is enabled.
   */
  def onAsyncStateCommitCompletion(event: AsyncStateCommitCompletionEvent): Unit = {
  }

  /**
   * Called when a query is stopped, with or without error.
   * @since 3.5.0
   */
  def onQueryTerminated(event: QueryTerminatedEvent): Unit
}

/**
 * Companion object of [[StreamingQueryListener]] that defines the listener events.
 * @since 3.5.0
 */
@Evolving
object StreamingQueryListener extends Serializable {

  /**
   * Base type of [[StreamingQueryListener]] events
   * @since 3.5.0
   */
  @Evolving
  trait Event

  /**
   * Event representing the start of a query
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param name
   *   User-specified name of the query, null if not specified.
   * @param timestamp
   *   The timestamp to start a query.
   * @since 3.5.0
   */
  @Evolving
  class QueryStartedEvent private[sql] (
      val id: UUID,
      val runId: UUID,
      val name: String,
      val timestamp: String)
      extends Event
      with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("name" -> JString(name)) ~
        ("timestamp" -> JString(timestamp))
    }
  }

  /**
   * Event representing any progress updates in a query.
   * @param progress
   *   The query progress updates.
   * @since 3.5.0
   */
  @Evolving
  class QueryProgressEvent private[sql] (val progress: StreamingQueryProgress)
      extends Event
      with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = JObject("progress" -> progress.jsonValue)
  }

  /**
   * Event representing that query is idle and waiting for new data to process.
   *
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param timestamp
   *   The timestamp when the latest no-batch trigger happened.
   * @since 3.5.0
   */
  @Evolving
  class QueryIdleEvent private[sql] (val id: UUID, val runId: UUID, val timestamp: String)
      extends Event
      with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("timestamp" -> JString(timestamp))
    }
  }

  /**
   * Event representing that termination of a query.
   *
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param exception
   *   The exception message of the query if the query was terminated with an exception.
   *   Otherwise, it will be `None`.
   * @param errorClassOnException
   *   The error class from the exception if the query was terminated with an exception which is a
   *   part of error class framework. If the query was terminated without an exception, or the
   *   exception is not a part of error class framework, it will be `None`.
   * @since 3.5.0
   */
  @Evolving
  class QueryTerminatedEvent private[sql] (
      val id: UUID,
      val runId: UUID,
      val exception: Option[String],
      val errorClassOnException: Option[String])
      extends Event
      with Serializable {
    // compatibility with versions in prior to 3.5.0
    def this(id: UUID, runId: UUID, exception: Option[String]) = {
      this(id, runId, exception, None)
    }

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("exception" -> JString(exception.orNull)) ~
        ("errorClassOnException" -> JString(errorClassOnException.orNull))
    }
  }

  /**
   * Event representing the async state commit completion for a microbatch.
   * @param id A unique query id that persists across restarts. See [[StreamingQuery.id()]].
   * @param runId A query id that is unique for every start/restart. See [[StreamingQuery.runId()]].
   * @param batchId Id of the microbatch this state commit belongs to.
   * @param stateOperatorCommitInfo One entry for each state operator in the query plan.
   *                                Empty in case of a failed commit.
   * @param exception The exception message of the commit if the commit failed with an error.
   *                  Otherwise, it will be `None`
   */
  @Evolving
  class AsyncStateCommitCompletionEvent private[sql](
      val id: UUID,
      val runId: UUID,
      val batchId: Long,
      val stateOperatorCommitInfo: Array[StatefulOperatorCommitInfo],
      val exception: Option[String])
      extends Event
      with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("batchId" -> org.json4s.JLong(batchId)) ~
        ("stateOperatorCommitInfo" -> JArray(stateOperatorCommitInfo.map(_.jsonValue).toList)) ~
        ("exception" -> JString(exception.orNull))
    }
  }

  /**
   * Stateful operator commit info including metrics.
   *
   * @param operatorName Name of the stateful operator instance.
   * @param numShufflePartitions Number of shuffle partitions this state operator is part of.
   * @param numStateStoreInstances Total number of state stores in the operator across all tasks.
   * @param customMetrics Custom metrics depending upon the state store. It is a map containing
   *                      metric name to metric value that is aggregated(sum) across the state
   *                      operator in all partitions. Currently all metric values are of long type.
   *                      Ex. Map("metric1" -> 25L, "metric2" -> 46L, "metric3" -> 9L). The metrics
   *                      depend on the type of store implementation used.
   */
  @Evolving
  class StatefulOperatorCommitInfo(
      val operatorName: String,
      val numShufflePartitions: Long,
      val numStateStoreInstances: Long,
      val customMetrics: java.util.Map[String, JLong]) extends Serializable {
    /** The compact JSON representation of this progress. */
    def json: String = compact(render(jsonValue))

    /** The pretty (i.e. indented) JSON representation of this progress. */
    def prettyJson: String = pretty(render(jsonValue))

    private[sql] def jsonValue: JValue = {
      ("operatorName" -> operatorName) ~
        ("numShufflePartitions" -> org.json4s.JLong(numShufflePartitions)) ~
        ("numStateStoreInstances" -> org.json4s.JLong(numStateStoreInstances)) ~
        ("customMetrics" -> {
          if (!customMetrics.isEmpty) {
            val keys = customMetrics.keySet.asScala.toSeq.sorted
            keys.map { k => k -> org.json4s.JLong(customMetrics.get(k)) : JObject }.reduce(_ ~ _)
          } else {
            JNothing
          }
        })
    }

    override def toString: String = prettyJson
  }
}

==> databricks-connect-15.4.0-sources.jar: org/apache/spark/sql/streaming/StreamingQueryListener.scala <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql.streaming

import java.lang.{Long => JLong}
import java.util.UUID

import scala.collection.JavaConverters._

import org.json4s._
import org.json4s.JsonAST.JValue
import org.json4s.JsonDSL._
import org.json4s.jackson.JsonMethods._

import org.apache.spark.annotation.Evolving

/**
 * Interface for listening to events related to [[StreamingQuery StreamingQueries]].
 * @note
 *   The methods are not thread-safe as they may be called from different threads.
 *
 * @since 3.5.0
 */
@Evolving
abstract class StreamingQueryListener extends Serializable {

  import StreamingQueryListener._

  /**
   * Called when a query is started.
   * @note
   *   This is called synchronously with
   *   [[org.apache.spark.sql.streaming.DataStreamWriter `DataStreamWriter.start()`]], that is,
   *   `onQueryStart` will be called on all listeners before `DataStreamWriter.start()` returns
   *   the corresponding [[StreamingQuery]]. Please don't block this method as it will block your
   *   query.
   * @since 3.5.0
   */
  def onQueryStarted(event: QueryStartedEvent): Unit

  /**
   * Called when there is some status update (ingestion rate updated, etc.)
   *
   * @note
   *   This method is asynchronous. The status in [[StreamingQuery]] will always be latest no
   *   matter when this method is called. Therefore, the status of [[StreamingQuery]] may be
   *   changed before/when you process the event. E.g., you may find [[StreamingQuery]] is
   *   terminated when you are processing `QueryProgressEvent`.
   * @since 3.5.0
   */
  def onQueryProgress(event: QueryProgressEvent): Unit

  /**
   * Called when the query is idle and waiting for new data to process.
   * @since 3.5.0
   */
  def onQueryIdle(event: QueryIdleEvent): Unit = {}

  /**
   * Called when there is an asynchronous state commit complete for a microbatch in the query.
   * This event is generated only when asynchronous state checkpointing is enabled.
   */
  def onAsyncStateCommitCompletion(event: AsyncStateCommitCompletionEvent): Unit = {}

  /**
   * Called when a query is stopped, with or without error.
   * @since 3.5.0
   */
  def onQueryTerminated(event: QueryTerminatedEvent): Unit
}

/**
 * Companion object of [[StreamingQueryListener]] that defines the listener events.
 * @since 3.5.0
 */
@Evolving
object StreamingQueryListener extends Serializable {

  /**
   * Base type of [[StreamingQueryListener]] events
   * @since 3.5.0
   */
  @Evolving
  trait Event

  /**
   * Event representing the start of a query
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param name
   *   User-specified name of the query, null if not specified.
   * @param timestamp
   *   The timestamp to start a query.
   * @since 3.5.0
   */
  @Evolving
  class QueryStartedEvent private[sql] (
      val id: UUID,
      val runId: UUID,
      val name: String,
      val timestamp: String)
    extends Event
    with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("name" -> JString(name)) ~
        ("timestamp" -> JString(timestamp))
    }
  }

  /**
   * Event representing any progress updates in a query.
   * @param progress
   *   The query progress updates.
   * @since 3.5.0
   */
  @Evolving
  class QueryProgressEvent private[sql] (val progress: StreamingQueryProgress)
    extends Event
    with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = JObject("progress" -> progress.jsonValue)
  }

  /**
   * Event representing that query is idle and waiting for new data to process.
   *
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param timestamp
   *   The timestamp when the latest no-batch trigger happened.
   * @since 3.5.0
   */
  @Evolving
  class QueryIdleEvent private[sql] (val id: UUID, val runId: UUID, val timestamp: String)
    extends Event
    with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("timestamp" -> JString(timestamp))
    }
  }

  /**
   * Event representing that termination of a query.
   *
   * @param id
   *   A unique query id that persists across restarts. See `StreamingQuery.id()`.
   * @param runId
   *   A query id that is unique for every start/restart. See `StreamingQuery.runId()`.
   * @param exception
   *   The exception message of the query if the query was terminated with an exception.
   *   Otherwise, it will be `None`.
   * @param errorClassOnException
   *   The error class from the exception if the query was terminated with an exception which is a
   *   part of error class framework. If the query was terminated without an exception, or the
   *   exception is not a part of error class framework, it will be `None`.
   * @since 3.5.0
   */
  @Evolving
  class QueryTerminatedEvent private[sql] (
      val id: UUID,
      val runId: UUID,
      val exception: Option[String],
      val errorClassOnException: Option[String])
    extends Event
    with Serializable {
    // compatibility with versions in prior to 3.5.0
    def this(id: UUID, runId: UUID, exception: Option[String]) = {
      this(id, runId, exception, None)
    }

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("exception" -> JString(exception.orNull)) ~
        ("errorClassOnException" -> JString(errorClassOnException.orNull))
    }
  }

  /**
   * Event representing the async state commit completion for a microbatch.
   * @param id
   *   A unique query id that persists across restarts. See [[StreamingQuery.id()]].
   * @param runId
   *   A query id that is unique for every start/restart. See [[StreamingQuery.runId()]].
   * @param batchId
   *   Id of the microbatch this state commit belongs to.
   * @param stateOperatorCommitInfo
   *   One entry for each state operator in the query plan. Empty in case of a failed commit.
   * @param exception
   *   The exception message of the commit if the commit failed with an error. Otherwise, it will
   *   be `None`
   */
  @Evolving
  class AsyncStateCommitCompletionEvent private[sql] (
      val id: UUID,
      val runId: UUID,
      val batchId: Long,
      val stateOperatorCommitInfo: Array[StatefulOperatorCommitInfo],
      val exception: Option[String])
    extends Event
    with Serializable {

    def json: String = compact(render(jsonValue))

    private def jsonValue: JValue = {
      ("id" -> JString(id.toString)) ~
        ("runId" -> JString(runId.toString)) ~
        ("batchId" -> org.json4s.JLong(batchId)) ~
        ("stateOperatorCommitInfo" -> JArray(stateOperatorCommitInfo.map(_.jsonValue).toList)) ~
        ("exception" -> JString(exception.orNull))
    }
  }

  /**
   * Stateful operator commit info including metrics.
   *
   * @param operatorName
   *   Name of the stateful operator instance.
   * @param numShufflePartitions
   *   Number of shuffle partitions this state operator is part of.
   * @param numStateStoreInstances
   *   Total number of state stores in the operator across all tasks.
   * @param customMetrics
   *   Custom metrics depending upon the state store. It is a map containing metric name to metric
   *   value that is aggregated(sum) across the state operator in all partitions. Currently all
   *   metric values are of long type. Ex. Map("metric1" -> 25L, "metric2" -> 46L, "metric3" ->
   *   9L). The metrics depend on the type of store implementation used.
   */
  @Evolving
  class StatefulOperatorCommitInfo(
      val operatorName: String,
      val numShufflePartitions: Long,
      val numStateStoreInstances: Long,
      val customMetrics: java.util.Map[String, JLong])
    extends Serializable {

    /** The compact JSON representation of this progress. */
    def json: String = compact(render(jsonValue))

    /** The pretty (i.e. indented) JSON representation of this progress. */
    def prettyJson: String = pretty(render(jsonValue))

    private[sql] def jsonValue: JValue = {
      ("operatorName" -> operatorName) ~
        ("numShufflePartitions" -> org.json4s.JLong(numShufflePartitions)) ~
        ("numStateStoreInstances" -> org.json4s.JLong(numStateStoreInstances)) ~
        ("customMetrics" -> {
          if (!customMetrics.isEmpty) {
            val keys = customMetrics.keySet.asScala.toSeq.sorted
            keys.map { k => k -> org.json4s.JLong(customMetrics.get(k)): JObject }.reduce(_ ~ _)
          } else {
            JNothing
          }
        })
    }

    override def toString: String = prettyJson
  }
}
